import json
import os
import re
import warnings


def readfile_or_default(fp, default=''):
    if os.path.exists(fp):
        with open(fp) as fd:
            return fd.read()
    return default


def count_slashes(input_string):
    dot_count = 0
    for char in input_string:
        if char == '/':
            dot_count += 1
    return dot_count


def is_name_legit(input_string):
    return bool(re.match(r'^(?=[a-zA-Z])[a-zA-Z0-9_-]*$', input_string))


def parse_dataset_name(name: str):
    s_count = count_slashes(name)
    if s_count == 1:
        username, dataset_name = name.split('/')
    else:
        raise ValueError('dataset name must be [username]/[dataset_name]')

    if not is_name_legit(username) or not is_name_legit(dataset_name):
        raise ValueError('dataset name and store name must startswith a-zA-Z, and contains only [a-zA-Z0-9_-]')

    return username, dataset_name


def read_jsonl(line_stream):
    for line in line_stream:
        try:
            yield json.loads(line)
        except json.decoder.JSONDecodeError:
            pass


def id_function(x):
    return x


def get_tqdm(silent=False):
    if silent:
        return id_function
    try:
        import tqdm
        return tqdm.tqdm
    except ModuleNotFoundError as e:
        warnings.warn('cannot load module tqdm, maybe it is not installed ?', UserWarning)
        return id_function


def parse_tags(tag: str) -> list[str]:
    return [t.strip() for t in tag.split(',') if t.strip() != '']



import time
from enum import Enum
from typing import Optional, List

from pydantic import BaseModel, Field

from .typing.core import DatasetSchema


class BinaryFile(BaseModel):
    name: str
    source: Optional[str] = None


class DataItemStorageFormat(str, Enum):
    jsonl = 'jsonl'


class CollectionConfig(BaseModel):
    name: str
    data_format: DataItemStorageFormat = DataItemStorageFormat.jsonl
    data_schema: DatasetSchema


class DatasetFileMeta(BaseModel):
    author: Optional[str] = None
    authorEmail: Optional[str] = None
    description: Optional[str] = None
    tags: Optional[List[str]] = None
    collections: List[CollectionConfig] = Field(default_factory=list)

    def find_collection(self, name) -> Optional[CollectionConfig]:
        matched_coll = [c for c in self.collections if c.name == name]
        if len(matched_coll) == 0:
            return None
        return matched_coll[0]



from datetime import date, datetime, time, timedelta

from dataset_sh.typing.core import FieldType, PydanticFieldDefinition, ClassDefinition, \
    DatasetSchema


def create_optional_field_type(name) -> FieldType:
    return FieldType.typed(
        'typing.Optional',
        [
            FieldType.simple(name)
        ]
    )


def guess_data_type(v):
    if isinstance(v, int):
        return 'int'
    elif isinstance(v, float):
        return 'float'
    elif isinstance(v, bool):
        return 'bool'
    elif isinstance(v, str):
        return 'str'
    elif isinstance(v, list):
        return 'list'
    elif isinstance(v, dict):
        return 'dict'
    elif isinstance(v, tuple):
        return 'tuple'
    elif isinstance(v, set):
        return 'set'
    elif isinstance(v, date):
        return 'datetime.date'
    elif isinstance(v, datetime):
        return 'datetime.datetime'
    elif isinstance(v, time):
        return 'datetime.time'
    elif isinstance(v, timedelta):
        return 'datetime.timedelta'
    else:
        return 'typing.Any'


def json_list_to_pydantic_code(items, max_row=None):
    clz_name = 'Model'
    field_map = {}
    for i, row in enumerate(items):
        if max_row is not None and i > max_row:
            break
        for col_name, cell in row.items():
            cell_type = guess_data_type(cell)
            if col_name in field_map:
                if field_map[col_name] != cell_type:
                    cell_type = 'typing.Any'
            field_map[col_name] = cell_type
    return DatasetSchema(
        entry_point=clz_name,
        classes=[
            ClassDefinition(
                name=clz_name,
                fields=[
                    PydanticFieldDefinition(
                        name=k,
                        v=create_optional_field_type(v)
                    )
                    for k, v in
                    field_map.items()]
            )
        ]
    )



import json
import os.path
import warnings
from typing import List, Optional
from pydantic import BaseModel, Field, ValidationError

from dataset_sh.server.auth import decode_key, verify_password, hash_password, encode_key, generate_password


class RepoServerUserProfile(BaseModel):
    user: str
    hashed: str


class RepoServerConfig(BaseModel):
    require_auth: bool = False

    allow_upload: bool = False

    hostname: str = 'http://127.0.0.1:5000'
    users: List[RepoServerUserProfile] = Field(default_factory=list)
    secret: str = Field(default_factory=generate_password)
    data_folder: str = os.path.expanduser('~/dataset_sh/storage')
    uploader_folder: str = os.path.expanduser('~/dataset_sh/uploader')
    article_folder: str = os.path.expanduser('~/dataset_sh/posts')

    def override_from_env(self):
        self.data_folder = os.getenv('DATASET_APP_STORAGE_LOCATION', self.data_folder)
        self.uploader_folder = os.getenv('DATASET_APP_UPLOADER_DIR', self.uploader_folder)
        self.article_folder = os.getenv('DATASET_APP_ARTICLE_LOCATION', self.article_folder)

        return self

    def make_dirs(self):
        os.makedirs(self.data_folder, exist_ok=True)
        os.makedirs(self.uploader_folder, exist_ok=True)
        os.makedirs(self.article_folder, exist_ok=True)

    @staticmethod
    def load_from_file(fp):
        try:
            with open(fp, 'r') as fin:
                return RepoServerConfig(**json.load(fin))
        except (FileNotFoundError, json.JSONDecodeError, ValidationError):
            warnings.warn('cannot load server config, using default config now ')
            return RepoServerConfig()

    def write_to_file(self, fp):
        with open(fp, 'w') as out:
            json.dump(self.model_dump(mode='json'), out, indent=4)

    @staticmethod
    def generate_key(username, password):
        return encode_key(username, password)

    def update_password(self, username, password):
        for u in self.users:
            if u.user == username:
                u.hashed = hash_password(password)
                return
        self.users.append(RepoServerUserProfile(
            user=username,
            hashed=hash_password(password)
        ))

    def verify_key(self, key) -> Optional[str]:
        username, password = decode_key(key)
        if self.verify_userpass(username, password):
            return username
        return None

    def verify_userpass(self, username, password) -> bool:
        u = self.get_user(username)
        if u:
            return verify_password(password, u.hashed)
        else:
            return False

    def get_user(self, username):
        for u in self.users:
            if u.user == username:
                return u
        return None



import os.path

CONFIG_JSON = os.path.expanduser('~/.dataset_sh_config.json')
STORAGE_BASE = os.path.expanduser('~/dataset_sh/storage')
ALIAS_FILE = os.path.expanduser('~/.dataset.sh.host-alias.json')
CENTRAL_HOST = 'https://remote.dataset.sh'



import json
import os
import warnings
import zipfile
from typing import Optional, Type, TypeVar, List

from pydantic import BaseModel

from .models import DatasetFileInternalPath
from .typing.codegen import CodeGenerator
from .typing.schema_builder import SchemaBuilder
from .utils.misc import id_function
from .core import CollectionConfig, DatasetFileMeta
from .utils.sample import reservoir_sampling

DataModel = TypeVar('DataModel', bound=BaseModel)


class DatasetFile:

    def __init__(self):
        raise ValueError('Please use DatasetFile.open(filename, mode)')

    @staticmethod
    def open(fp: str, mode: str = 'r'):
        """
        Open a dataset file
        :param fp: path to the file
        :param mode: r for read and w for write.
        :return:
        """
        if mode == 'r':
            return DatasetFileReader(fp)
        elif mode == 'w':
            return DatasetFileWriter(fp)
        else:
            raise ValueError('mode must be one of "r" or "w"')

    @staticmethod
    def binary_file_path(fn: str):
        return os.path.join(DatasetFileInternalPath.BINARY_FOLDER, fn)


class DatasetFileWriter:
    def __init__(self, file_path: str, compression=zipfile.ZIP_LZMA, compresslevel=9):
        """
        Write to a dataset file, this object can also be used as a context manager.

        This object need to be closed.

        :param file_path: location of the dataset file to write.
        :param compression: compress mode for zip file.
        :param compresslevel: note that the default compression algorithm ZIP_LZMA do not use this value.

        """
        self.zip_file = zipfile.ZipFile(file_path, 'w', compression=compression, compresslevel=compresslevel)
        self.meta = DatasetFileMeta()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        """
        close the writer.
        :return:
        """
        with self.zip_file.open(DatasetFileInternalPath.META_FILE_NAME, 'w') as out:
            out.write(self.meta.model_dump_json().encode('utf-8'))
        self.zip_file.close()

    def add_collection(
            self,
            collection_name: str,
            data: List,
            model: Optional[Type[DataModel]] = None,
            tqdm=id_function,
    ):
        """
        add a data collection to this dataset.
        :param collection_name: name of the collection to add.
        :param data: list of objects that extends pydantic BaseModel.
        :param model: the pydantic model class.
        :param tqdm: Optional tqdm progress bar.
        :return:
        """
        for coll in self.meta.collections:
            if coll.name == collection_name:
                raise ValueError(f'collection {collection_name} already exists')

        if model is None:
            if isinstance(data[0], BaseModel):
                warnings.warn(f'model class is not provided, using {data[0].__class__.__name__} as model')
                model = data[0].__class__
            else:
                raise ValueError('Input data must be pydantic model')

        dataset_schema = SchemaBuilder.build(model)

        new_coll = CollectionConfig(
            name=collection_name,
            data_schema=dataset_schema,
        )

        self.meta.collections.append(new_coll)
        target_fp = os.path.join(
            DatasetFileInternalPath.COLLECTION_FOLDER,
            collection_name,
            DatasetFileInternalPath.DATA_FILE
        )
        with self.zip_file.open(target_fp, 'w') as out:
            for item in tqdm(data):
                out.write(item.model_dump_json(round_trip=True).encode('utf-8'))
                out.write("\n".encode('utf-8'))

    def add_binary_file(self, fn: str, content: bytes):
        """
        Add a binary file to the dataset
        :param fn: name of the binary file.
        :param content: content in bytes.
        :return:
        """
        binary_file_path = DatasetFile.binary_file_path(fn)
        with self.zip_file.open(binary_file_path, 'w') as out:
            out.write(content)


class DatasetFileReader:
    def __init__(self, file_path):
        """
        Read a dataset, this object can be used as a context manager.

        This object must be closed.

        :param file_path:
        """
        self.zip_file = zipfile.ZipFile(file_path, 'r')

        with self.zip_file.open(DatasetFileInternalPath.META_FILE_NAME, 'r') as fd:
            self.meta = DatasetFileMeta(**json.load(fd))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        self.zip_file.close()

    def binary_files(self):
        """
        Open a binary file for read.
        :return: a file descriptor for the binary file to read.
        """
        prefix = DatasetFileInternalPath.BINARY_FOLDER + '/'
        for name in self.zip_file.namelist():
            if name.startswith(prefix):
                yield name[len(prefix):]

    def open_binary_file(self, filename):
        """
        Open a binary file for read.
        :param filename: name of the binary file.
        :return: a file descriptor for the binary file to read.
        """
        return self.zip_file.open(
            DatasetFile.binary_file_path(filename),
            'r'
        )

    def collection(self, collection_name, model=None):
        """
        Open a collection.
        :param collection_name: name of a collection
        :param model: an optional pydantic model class to hold the data.
        :return: a CollectionReader object for the given collection name.
        """
        cfg = [c for c in self.meta.collections if c.name == collection_name]
        if len(cfg) == 0:
            raise ValueError(f"Collection {collection_name} do not exist")
        else:
            cfg = cfg[0]
        return CollectionReader(self.zip_file, collection_name, cfg, model=model)

    def coll(self, collection_name, model=None):
        return self.collection(collection_name, model=model)

    def collections(self):
        """
        List all collection names
        :return: list of collection names.
        """
        return [c.name for c in self.meta.collections]

    def __getitem__(self, item):
        return self.collection(item)


class CollectionReader(object):
    def __init__(self, zip_file, collection_name, config: CollectionConfig, model=None):
        """
        Collection Reader
        :param zip_file:
        :param collection_name:
        :param config:
        :param model:
        """
        self.zip_file = zip_file
        self.collection_name = collection_name
        self.config = config
        self.model = model

    def code_usage(self):
        generator = CodeGenerator()
        code = generator.generate_all(self.config.data_schema)
        return code

    def top(self, n=10):
        ret = []
        for i, row in enumerate(self):
            if i >= n:
                break
            ret.append(row)
        return ret

    def random_sample(self, n=10):
        return reservoir_sampling(self, n)

    def __iter__(self):
        """
        Iterate through the collection.
        :return:
        """
        entry = os.path.join(
            DatasetFileInternalPath.COLLECTION_FOLDER,
            self.collection_name,
            DatasetFileInternalPath.DATA_FILE
        )
        with self.zip_file.open(entry, 'r') as fd:
            for line in fd:
                line = line.strip()
                if len(line) > 0:
                    item = json.loads(line)
                    if self.model is None:
                        yield item
                    else:
                        yield self.model(**item)

    def to_list(self):
        """
        Read the collection as list instead of iterator
        :return:
        """
        return list(self)


# Standard IO operations


def open_dataset_file(fp: str) -> 'DatasetFileReader':
    """
    Read a dataset file
    :param fp: path to the file
    :return:
    """
    return DatasetFileReader(fp)


def create(fp: str, compression=zipfile.ZIP_LZMA, compresslevel=9) -> 'DatasetFileWriter':
    """
    Create a dataset file to write
    :param fp: path to the file
    :param compression: compress mode for zip file.
    :param compresslevel: note that the default compression algorithm ZIP_LZMA do not use this value.
    :return:
    """
    return DatasetFileWriter(fp, compression=compression, compresslevel=compresslevel)



import json
import os
import warnings

from dataset_sh.server.core import RepoServerConfig


def load_config(app=None) -> RepoServerConfig:
    config_file = os.environ.get('DATASET_SH_SERVER_CONFIG_FILE', './dataset-sh-server-config.json')

    if config_file is None or not os.path.exists(config_file):
        if not os.path.exists(config_file):
            warnings.warn('server config do not exist, server will be using a default configuration file, '
                          'which may affect some functionality, consider create and modify a configuration file.')
        host = 'localhost'
        port = '5000'
        if app is not None:
            host = app.config.get('SERVER_NAME', host)
            if host is None:
                host = '127.0.0.1'
            port = app.config.get('SERVER_PORT', port)
        return RepoServerConfig(hostname=f'http://{host}:{port}').override_from_env()
    else:
        with open(config_file) as fd:
            return RepoServerConfig(**json.load(fd)).override_from_env()



from __future__ import annotations
import json
import os.path
import shutil
import tempfile
import uuid
import warnings
from typing import Optional, List

import requests

from dataset_sh import DatasetFile, open_dataset_file
from dataset_sh.clients.common import MarkedFolder
from dataset_sh.clients.remote import url_join, TagResolveResult, ListVersionResult

from dataset_sh.core import DatasetFileMeta
from dataset_sh.models import DatasetFileInternalPath
from dataset_sh.remote_utils.profile import DatasetClientProfileConfig
from dataset_sh.typing.codegen import CodeGenerator
from dataset_sh.utils.files import checksum, upload_to_url, download_url
from dataset_sh.utils.misc import parse_dataset_name, read_jsonl
import dataset_sh.constants as DatasetConstants


class LocalDatasetVersion(MarkedFolder):
    location: str
    namespace: str
    dataset_name: str
    version: str

    def __init__(self, location, parent, namespace, dataset_name, version):
        super().__init__(location, parent)
        self.namespace = namespace
        self.dataset_name = dataset_name
        self.version = version

    def __eq__(self, other):
        return self.namespace == other.namespace and self.dataset_name == other.dataset_name and self.version == other.version

    def open(self):
        return DatasetFile.open(self.datafile(), 'r')

    # Filepaths
    def marker_file(self) -> str:
        return os.path.join(self.location, '.dataset.version.marker')

    def datafile(self):
        return os.path.join(self.location, 'file')

    def metafile(self):
        return os.path.join(self.location, 'meta.json')

    def code_example_file(self, collection_name):
        return os.path.join(self.location, f'usage_code_{collection_name}.py')

    def sample_file(self, collection_name):
        return os.path.join(self.location, f'data_sample_{collection_name}.jsonl')

    # End of Filepaths

    def __repr__(self):
        return f'{self.namespace}/{self.dataset_name}:version={self.version}'

    def __str__(self):
        return f'{self.namespace}/{self.dataset_name}:version={self.version}'

    def meta(self) -> DatasetFileMeta | dict:
        with open(self.metafile()) as fd:
            return json.load(fd)

    def meta_object(self) -> DatasetFileMeta:
        return DatasetFileMeta(**self.meta())

    def sample(self, collection):
        if os.path.isfile(self.sample_file(collection)):
            with open(self.sample_file(collection)) as fd:
                return list(read_jsonl(fd))
        return []

    def usage_code(self, collection_name):
        meta = self.meta()
        meta = DatasetFileMeta(**meta)
        cs = [c for c in meta.collections if c.name == collection_name]
        if len(cs) > 0:
            cg = CodeGenerator()
            return cg.generate_all(cs[0].data_schema)
        else:
            return ''

    def delete(self):
        shutil.rmtree(self.location)

    def upload_to(self, dest: 'RemoteDataset', tags: Optional[List[str]] = None):
        dest.upload_from_local(self, tags=tags)

    def extract_sample_and_code_usage(self):
        with open_dataset_file(self.datafile()) as reader:
            meta_file_dest = self.metafile()
            with reader.zip_file.open(DatasetFileInternalPath.META_FILE_NAME, 'r') as fd:
                with open(meta_file_dest, 'wb') as out:
                    out.write(fd.read())

            for coll_name in reader.collections():
                coll = reader.collection(coll_name)
                with open(self.sample_file(coll_name), 'w') as out:
                    samples = coll.top(n=10)
                    for item in samples:
                        out.write(f'{json.dumps(item)}\n')
                with open(self.code_example_file(coll_name), 'w') as out:
                    code = coll.code_usage()
                    out.write(code)


class LocalDataset(MarkedFolder):
    location: str
    namespace: str
    dataset_name: str

    def __init__(self, location, parent, namespace, dataset_name):
        super().__init__(location, parent)
        self.namespace = namespace
        self.dataset_name = dataset_name

    def __eq__(self, other):
        return self.namespace == other.namespace and self.dataset_name == other.dataset_name

    def delete(self):
        shutil.rmtree(self.location)

    def post_create(self):
        os.makedirs(os.path.join(self.location, 'version'), exist_ok=True)

    def marker_file(self):
        return os.path.join(self.location, '.dataset.marker')

    def version_folder(self):
        return os.path.join(self.location, 'version')

    def tag_file(self):
        return os.path.join(self.location, 'tag')

    def read_tag_file(self):
        if os.path.exists(self.tag_file()):
            with open(self.tag_file()) as fd:
                return json.load(fd)
        else:
            return {"tags": {}}

    def save_tag_file(self, tags):
        with open(self.tag_file(), 'w') as fd:
            json.dump(tags, fd)
        return self

    def set_tag(self, tag: str, version: str):
        tags = self.read_tag_file()
        tags['tags'][tag] = version
        self.save_tag_file(tags)
        return self

    def remove_tag(self, tag):
        tags = self.read_tag_file()
        del tags['tags'][tag]
        self.save_tag_file(tags)
        return self

    def resolve_tag(self, tag) -> Optional[str]:
        tags = self.read_tag_file()
        return tags['tags'].get(tag, None)

    def versions(self) -> list[LocalDatasetVersion]:
        items = []
        for version_id in os.listdir(self.version_folder()):
            v = self.version(version_id)
            if v.exists():
                items.append(v)
        return items

    def version(self, version_id) -> LocalDatasetVersion:
        version_location = os.path.join(self.location, 'version', version_id)
        return LocalDatasetVersion(version_location, self, self.namespace, self.dataset_name, version_id)

    def tags(self):
        tags = self.read_tag_file()
        return tags['tags']

    def tag(self, tag, allow_none=False) -> LocalDatasetVersion | None:
        version = self.resolve_tag(tag)
        if version is None:
            if allow_none:
                return None
            else:
                raise ValueError('tag do not exist.')
        return self.version(version)

    def latest(self) -> LocalDatasetVersion | None:
        return self.tag('latest')

    def open(self):
        return self.latest().open()

    def import_file(
            self,
            file_path,
            replace=False,
            remove_source=False,
            verify_version=None,
            tags: Optional[List[str]] = None,
            as_latest=True
    ) -> LocalDatasetVersion | None:
        new_version_id = checksum(file_path)

        if verify_version and verify_version != new_version_id:
            raise ValueError('provided version do not match file signature.')
        new_version = self.version(new_version_id)
        if new_version.exists():
            if not replace:
                warnings.warn(f'dataset {self.namespace}/{self.dataset_name} version {new_version_id} already exists')
                return

        new_version_folder = new_version
        os.makedirs(
            new_version_folder.location, exist_ok=False
        )

        if remove_source:
            shutil.move(file_path, new_version_folder.datafile())
        else:
            shutil.copy2(file_path, new_version_folder.datafile())

        new_version.create()
        new_version.extract_sample_and_code_usage()
        if tags:
            for t in tags:
                self.set_tag(t, new_version_id)

        if as_latest:
            self.set_tag('latest', new_version_id)

        return new_version

    def __repr__(self):
        return f'{self.namespace}/{self.dataset_name}'

    def __str__(self):
        return f'{self.namespace}/{self.dataset_name}'


class LocalNamespace(MarkedFolder):
    location: str
    namespace: str

    def __init__(self, location, parent, namespace):
        super().__init__(location, parent)
        self.namespace = namespace

    def __eq__(self, other):
        return self.namespace == other.namespace

    def marker_file(self):
        return os.path.join(self.location, '.dataset.namespace.marker')

    def datasets(self) -> list[LocalDataset]:
        datasets = []
        if os.path.exists(self.location):
            for dataset_name in os.listdir(self.location):
                if self.dataset(dataset_name).exists():
                    datasets.append(self.dataset(dataset_name))
        return datasets

    def dataset(self, dataset_name) -> LocalDataset:
        loc = os.path.join(self.location, dataset_name)
        return LocalDataset(loc, self, self.namespace, dataset_name)

    def __repr__(self):
        return f'Dataset Namespace: {self.namespace}'

    def __str__(self):
        return f'Dataset Namespace: {self.namespace}'


class LocalStorage(object):
    location: str

    def __init__(self, location=None):
        if location is None:
            location = DatasetConstants.STORAGE_BASE
        self.location = location
        os.makedirs(self.location, exist_ok=True)

    def namespaces(self) -> list[LocalNamespace]:
        namespaces = []
        if os.path.exists(self.location):
            for namespace in os.listdir(self.location):
                if self.namespace(namespace).exists():
                    namespaces.append(self.namespace(namespace))
        return namespaces

    def namespace(self, name) -> LocalNamespace:
        loc = os.path.join(
            self.location,
            name,
        )
        return LocalNamespace(loc, parent=None, namespace=name)

    def datasets(self) -> list[LocalDataset]:
        datasets = []
        for namespace in self.namespaces():
            for d in namespace.datasets():
                datasets.append(d)
        return datasets

    def dataset(self, name) -> LocalDataset:
        username, dataset_name = parse_dataset_name(name)
        return self.namespace(username).dataset(dataset_name)


class RemoteInfo:
    host: str
    access_key: Optional[str]

    def __init__(self, host, access_key=None):
        self.host = host
        self.access_key = access_key

    def get_headers(self):
        if self.access_key:
            return {"X-DATASET-SH-ACCESS-KEY": self.access_key}
        return None


class RemoteDatasetVersionTag(object):
    remote: RemoteInfo

    namespace: str
    dataset_name: str
    tag: str

    def __init__(self, remote, namespace, dataset_name, tag):
        self.remote = remote
        self.namespace = namespace
        self.dataset_name = dataset_name
        self.tag = tag

    def _get_download_url(self) -> str:
        api_url = url_join(self.remote.host, 'api', 'dataset', self.namespace, self.dataset_name, 'tag', self.tag,
                           'file')
        return api_url

    def download_to_file(self, filepath: str):
        url = self._get_download_url()
        download_url(url, filepath, self.remote.get_headers())

    def download_to(self, dest: LocalDataset, keep_tag=True, additional_tags: Optional[List[str]] = None):
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_file_path = os.path.join(temp_dir, str(uuid.uuid4()))
            self.download_to_file(temp_file_path)
            version = dest.import_file(temp_file_path, remove_source=True)
            if version:
                if additional_tags:
                    for tag in additional_tags:
                        dest.set_tag(tag, version.version)
                if keep_tag:
                    dest.set_tag(self.tag, version.version)

    def download(self):
        self.download_to(LocalStorage().namespace(self.namespace).dataset(self.dataset_name), keep_tag=True)


class RemoteDatasetVersion(object):
    remote: RemoteInfo

    namespace: str
    dataset_name: str
    version: str

    def __init__(self, remote, namespace, dataset_name, version):
        self.remote = remote
        self.namespace = namespace
        self.dataset_name = dataset_name
        self.version = version

    def _get_download_url(self) -> str:
        api_url = url_join(self.remote.host, 'api', 'dataset', self.namespace, self.dataset_name, 'version',
                           self.version,
                           'file')
        return api_url

    def download_to_file(self, filepath: str):
        url = self._get_download_url()
        download_url(url, filepath, self.remote.get_headers())

    def download_to(self, dest: LocalDataset, tags: Optional[List[str]] = None):
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_file_path = os.path.join(temp_dir, str(uuid.uuid4()))
            self.download_to_file(temp_file_path)
            dest.import_file(temp_file_path, remove_source=True, verify_version=self.version)
            if tags:
                for tag in tags:
                    dest.set_tag(tag, self.version)

    def download(self):
        self.download_to(LocalStorage().namespace(self.namespace).dataset(self.dataset_name))


class RemoteDataset(object):
    remote: RemoteInfo

    namespace: str
    dataset_name: str

    def __init__(self, remote, namespace, dataset_name):
        self.remote = remote
        self.namespace = namespace
        self.dataset_name = dataset_name

    def _get_upload_url(self) -> str:
        api_url = url_join(self.remote.host, 'api', 'dataset', self.namespace, self.dataset_name)
        return api_url

    def resolve_tag(self, tag):
        api_url = url_join(self.remote.host, 'api', 'dataset', self.namespace, self.dataset_name, 'tag', tag)
        headers = self.remote.get_headers()
        resp = requests.get(
            api_url,
            headers=headers,
        )
        resp.raise_for_status()
        return TagResolveResult(**resp.json())

    def versions(self) -> ListVersionResult:
        api_url = url_join(self.remote.host, 'api', 'dataset', self.namespace, self.dataset_name, 'version')
        headers = self.remote.get_headers()
        resp = requests.get(
            api_url,
            headers=headers,
        )
        resp.raise_for_status()
        return ListVersionResult(**resp.json())

    def version(self, version_id) -> RemoteDatasetVersion:
        return RemoteDatasetVersion(self.remote, self.namespace, self.dataset_name, version_id)

    def tag(self, tag) -> RemoteDatasetVersionTag:
        return RemoteDatasetVersionTag(self.remote, self.namespace, self.dataset_name, tag)

    def latest(self) -> RemoteDatasetVersionTag:
        return self.tag('latest')

    def set_tag(self, tag, version):
        api_url = url_join(self.remote.host, 'api', 'dataset', self.namespace, self.dataset_name, 'tag', tag)
        headers = self.remote.get_headers()

        resp = requests.post(
            api_url,
            json={'version': version},
            headers=headers,
        )
        resp.raise_for_status()

    def upload_from_file(self, filepath, tags: Optional[List[str]] = None):
        upload_url = self._get_upload_url()
        params = None

        if tags:
            params['tag'] = ','.join(tags)

        upload_to_url(
            upload_url,
            filepath,
            self.remote.get_headers(),
            params,
            f'{self.namespace}/{self.dataset_name}'
        )

    def upload_from_local(self, source: LocalDatasetVersion, tags: Optional[List[str]] = None):
        self.upload_from_file(source.datafile(), tags=tags)


class RemoteNamespace(object):
    remote: RemoteInfo
    namespace: str

    def __init__(self, remote, namespace):
        self.remote = remote
        self.namespace = namespace

    def dataset(self, dataset_name) -> RemoteDataset:
        return RemoteDataset(self.remote, self.namespace, dataset_name)


class Remote(object):
    remote: RemoteInfo

    def __init__(self, host=None, access_key=None):
        if host is None:
            host = DatasetConstants.CENTRAL_HOST
        self.remote = RemoteInfo(host, access_key)

    def namespace(self, name) -> RemoteNamespace:
        return RemoteNamespace(self.remote, name)

    def dataset(self, name) -> RemoteDataset:
        username, dataset_name = parse_dataset_name(name)
        return RemoteDataset(self.remote, username, dataset_name)


def remote(host=None, profile=None, find_profile=True):
    """
    Create remote server based on:
        * host and find_profile
        * profile

    If profile is provided, dataset.sh will connect using the selected profile configuration.

    If host is provided:
        dataset.sh will connect to the given host address,
        and if find_profile is True, the first profile matching the given host url will be used.

    If nothing is provided,
        dataset.sh will connect using default profile, and if default profile is missing, it will connect to
        the default host as a visitor.

    Args:
        host:
        profile:
        find_profile:

    Returns:

    """
    if host is not None:
        if find_profile:
            profile = DatasetClientProfileConfig.load_profiles().get_by_host(host)
            if profile:
                return Remote(host, profile.key)
        return Remote(host)
    elif profile is not None:
        profile = DatasetClientProfileConfig.load_profiles().get_profile(profile_name=profile)
        return Remote(profile.host, profile.key)
    else:
        default_profile = DatasetClientProfileConfig.load_profiles().get_profile(profile_name='default')
        if default_profile is not None:
            return Remote(default_profile.host, default_profile.key)

        return Remote(DatasetConstants.CENTRAL_HOST)


def dataset(name):
    return LocalStorage().dataset(name)



import json

import click

from dataset_sh import open_dataset_file
from dataset_sh.typing.codegen import CodeGenerator


@click.group(name='file')
def file_cli():
    """to interact with standalone dataset.sh file"""
    pass


@file_cli.command(name='list-collections')
@click.argument('filepath', type=click.Path())
def list_collections(filepath):
    """
    list collections of a dataset
    """
    with open_dataset_file(filepath) as reader:
        collections = reader.collections()
        click.echo(f'Total collections: {len(collections)}')
        for coll in collections:
            click.echo(coll)


@file_cli.command(name='print-code')
@click.argument('filepath', type=click.Path())
@click.argument('collection_name')
def print_code(filepath, collection_name):
    with open_dataset_file(filepath) as reader:
        coll = reader.collection(collection_name)
        code = coll.code_usage()
        schema = coll.config.data_schema
        loader_code = CodeGenerator.generate_file_loader_code(filepath, collection_name, schema.entry_point)
        click.echo(code)
        click.echo('\n\n')
        click.echo(loader_code)


@file_cli.command(name='print-sample')
@click.argument('filepath', type=click.Path())
@click.argument('collection_name')
def print_sample(filepath, collection_name):
    with open_dataset_file(filepath) as reader:
        sample = reader.collection(collection_name).top()
        click.echo(json.dumps(sample, indent=2))



from typing import Union, Optional

from .core import EnumEntry, FieldType, PydanticFieldDefinition, EnumDefinition, ClassDefinition, \
    DatasetSchema
from .constants import SUPPORTED_TYPES


def doc_to_doc_str(doc_or_none: Optional[str]) -> Optional[str]:
    if doc_or_none is None:
        return None
    else:

        indented_doc_str = '\n'.join(
            [f"    {line}" for line in doc_or_none.split('\n')]
        )
        return "\n".join([
            '    """', indented_doc_str, '    """'
        ])


class CodeGenerator:
    def __init__(self):
        self.type_mapping = {}
        for t in SUPPORTED_TYPES:
            self.type_mapping[t] = t

    def field_type_to_code(self, ift: FieldType, current_class):
        n = self.type_mapping[ift.name]
        if ift.is_parametric:
            cs = ", ".join([self.field_type_to_code(c, current_class) for c in ift.children])
            return f"{n}[{cs}]"
        else:
            if n == current_class:
                return f"'{n}'"
            return f"{n}"

    def field_declaration_to_code(self, item: PydanticFieldDefinition, current_class):
        ts = self.field_type_to_code(item.type, current_class)
        if ts.strip().startswith('typing.Optional'):
            ts = f"{ts} = None"
        return f"{item.name} : {ts}"

    def enum_entry_to_code(self, e: EnumEntry):
        return f"{e.name} = {to_literal(e.value)}"

    def enum_def_to_code(self, item: EnumDefinition):
        tn = self.type_mapping[item.name]
        if item.is_int():
            header = f"class {tn}(IntEnum):"
        else:
            header = f"class {tn}(str, Enum):"
        body = '\n'.join([
            f"    {self.enum_entry_to_code(x)}" for x in item.enum_entries
        ])

        doc_str = doc_to_doc_str(item.doc_str)
        if doc_str is not None:
            return '\n'.join([header, doc_str, body])
        else:
            return '\n'.join([header, body])

    def class_def_to_code(self, item: ClassDefinition):
        tn = self.type_mapping[item.name]
        header = f"class {tn}(BaseModel):"
        body = '\n'.join([
            f"    {self.field_declaration_to_code(x, tn)}" for x in item.fields
        ])
        doc_str = doc_to_doc_str(item.doc_str)
        if doc_str is not None:
            return '\n'.join([header, doc_str, body])
        else:
            return '\n'.join([header, body])

    def class_or_enum_to_code(self, item: Union[ClassDefinition, EnumDefinition]):
        if isinstance(item, ClassDefinition):
            return self.class_def_to_code(item)
        elif isinstance(item, EnumDefinition):
            return self.enum_def_to_code(item)

    def generate_all(self, schema: DatasetSchema):
        # first rename class name if possible
        # generate class one by one based on topo order, which the classes
        # should already have been in that order if it is built by schema builder
        used_names = set()
        for k in self.type_mapping.values():
            used_names.add(k)

        for c in reversed(schema.classes):
            last_name = c.name.split('.')[-1]
            alter_name = last_name

            while alter_name in used_names:
                alter_name = f"{alter_name}_1"

            self.type_mapping[c.name] = alter_name
            used_names.add(alter_name)
        self.type_mapping = self.type_mapping
        header = 'import datetime\n' \
                 'import typing\n' \
                 'from enum import Enum, IntEnum\n' \
                 'import dataset_sh\n' \
                 'from pydantic import BaseModel\n'

        code_body = '\n\n'.join(
            [self.class_or_enum_to_code(c) for c in schema.classes]
        )

        return '\n'.join([
            header, code_body
        ])

    @staticmethod
    def generate_loader_code(store_name, dataset_name, collection_name, entry_class_name):
        entry_class_name = entry_class_name.split('.')[-1]
        code = f"""
with dataset_sh.read('{store_name}/{dataset_name}') as reader:
    for item in reader.coll('{collection_name}', model={entry_class_name}):
        print(item)
        break
"""
        return code

    @staticmethod
    def generate_file_loader_code(filepath, collection_name, entry_class_name):
        entry_class_name = entry_class_name.split('.')[-1]
        code = f"""
with dataset_sh.read_file('{filepath}') as reader:
    for item in reader.coll('{collection_name}', model={entry_class_name}):
        print(item)
        break
"""
        return code


def to_literal(v):
    if isinstance(v, str):
        return "'" + v.replace("'", "\\'") + "'"
    elif isinstance(v, int):
        return f"{v}"



from typing import Optional

from pydantic import BaseModel


#
# class HostAlias(BaseModel):
#     name: str
#     host: str
#
#
# class HostAliasList(BaseModel):
#     aliases: List[HostAlias] = Field(default_factory=list)
#
#     def resolve_alias(self, name) -> Optional[str]:
#         for item in self.aliases:
#             if item.name == name:
#                 return item.host
#         if name == 'default':
#             return DatasetConstants.CENTRAL_HOST
#         return None
#
#     def save(self):
#         config_file = DatasetConstants.ALIAS_FILE
#         data = self.model_dump(mode='json')
#         with open(config_file, 'w') as out:
#             json.dump(data, out, indent=4)
#
#     def add_alias(self, name, host):
#         HostAliasList.is_valid_host_address(host, throw=True)
#         for item in self.aliases:
#             if item.name == name:
#                 item.host = host
#                 return
#         self.aliases.append(HostAlias(name=name, host=host))
#         return self
#
#     def remove_alias(self, name):
#         aliases = []
#         for item in self.aliases:
#             if item.name != name:
#                 aliases.append(item)
#         self.aliases = aliases
#         return self
#
#     @staticmethod
#     def load_from_disk():
#         config_file = DatasetConstants.ALIAS_FILE
#         if os.path.exists(config_file):
#             with open(config_file, 'r') as content:
#                 try:
#                     json_content = json.load(content)
#                     ret = HostAliasList(**json_content)
#                     return ret
#                 except (ValidationError, json.decoder.JSONDecodeError):
#                     warnings.warn('cannot parse profile config')
#         return HostAliasList()
#
#     @staticmethod
#     def resolve_host_or_alias(host: Optional[str]):
#
#         if host is None:
#             host = 'default'
#
#         if '://' in host:
#             return host
#         host_addr = HostAliasList.load_from_disk().resolve_alias(host)
#         if host_addr:
#             if host_addr.startswith('http://') or host_addr.startswith('https://'):
#                 return host_addr
#             else:
#                 raise ValueError(f'Invalid host address: {host_addr}, it must starts with https:// or http://')

# @staticmethod
def is_valid_host_address(host_addr, throw=True):
    if host_addr.startswith('http://') or host_addr.startswith('https://'):
        return True
    else:
        if throw:
            raise ValueError(f'Invalid host address: {host_addr}, it must starts with https:// or http://')
        else:
            return False


class TagResolveResult(BaseModel):
    version: Optional[str] = None
    tag: str


class ListVersionResult(BaseModel):
    versions: list[str]


def url_join(base_url, *path_items):
    if base_url.endswith('/'):
        base_url = base_url[:-1]
    return '/'.join([base_url, *path_items])



from typing import Union, Optional, List

from pydantic import BaseModel


class EnumEntry(BaseModel):
    name: str
    value: Union[int, str]


class FieldType(BaseModel):
    is_parametric: bool
    name: str
    children: Optional[List['FieldType']] = None

    @staticmethod
    def simple(name):
        return FieldType(is_parametric=False, name=name)

    @staticmethod
    def typed(name, children):
        return FieldType(is_parametric=True, name=name, children=children)


class PydanticFieldDefinition(BaseModel):
    name: str
    type: FieldType


class EnumDefinition(BaseModel):
    name: str
    enum_entries: Optional[List[EnumEntry]]
    doc_str: Optional[str] = None

    def is_int(self):
        if self.enum_entries is not None and len(self.enum_entries) > 0:
            return isinstance(self.enum_entries[0].value, int)
        return False


class ClassDefinition(BaseModel):
    name: str
    fields: List[PydanticFieldDefinition]
    doc_str: Optional[str] = None


class DatasetSchema(BaseModel):
    entry_point: str
    classes: List[Union[EnumDefinition, ClassDefinition]]



from dataset_sh import create
from dataset_sh.utils.files import checksum
from dataset_sh.utils.misc import get_tqdm


def dump_single_collection(fn, name, data, silent=False):
    with create(fn) as out:
        out.add_collection(name, data, data[0].__class__, tqdm=get_tqdm(silent=silent))
    return checksum(fn)


def dump_collections(fn, data_dict, silent=False):
    tqdm = get_tqdm(silent=silent)
    with create(fn) as out:
        for name, data in tqdm(data_dict.items()):
            if len(data) > 0:
                out.add_collection(name, data, data[0].__class__)
    return checksum(fn)



import base64
import json
import secrets
import uuid

from werkzeug.security import generate_password_hash, check_password_hash


def generate_password():
    return secrets.token_hex(16)


def encode_key(username, password):
    key = json.dumps({
        'key_id': str(uuid.uuid4()),
        'username': username,
        'password': password
    })
    return base64.b64encode(key.encode("ascii")).decode('utf-8')


def decode_key(key):
    decoded = base64.b64decode(key).decode('utf-8')
    payload = json.loads(decoded)
    return payload['username'], payload['password']


def verify_password(password, hashed_password):
    return check_password_hash(hashed_password, password)


def hash_password(password):
    return generate_password_hash(password)



import hashlib
import json
import os
import warnings
from pathlib import Path

import requests


def checksum(file_path):
    md5_hash = hashlib.sha256()

    with open(file_path, "rb") as file:
        while True:
            data = file.read(65536)
            if not data:
                break
            md5_hash.update(data)

    return md5_hash.hexdigest()


def filesize(file_path):
    file_size = os.path.getsize(file_path)
    return file_size


def read_url_json(url, headers):  # pragma: no cover
    response = requests.get(url, headers=headers)
    response.raise_for_status()
    return response.json()


def download_url(url, target, headers):
    response = requests.get(url, headers=headers, stream=True)

    has_tqdm = True
    try:
        from tqdm import tqdm
    except ModuleNotFoundError:  # pragma: no cover
        has_tqdm = False

    if response.status_code == 200:
        with open(target, "wb") as file:
            if has_tqdm:
                file_size = int(response.headers.get('content-length', 0))
                desc = "(Unknown total file size)" if file_size == 0 else ""
                with tqdm(
                        total=file_size, unit='iB', unit_scale=True
                ) as bar:
                    for data in response.iter_content(chunk_size=1024):
                        size = file.write(data)
                        bar.update(size)
            else:
                for data in response.iter_content(chunk_size=1024):  # pragma: no cover
                    size = file.write(data)

        response.close()
    else:
        server_err = json.loads(response.content.decode('utf-8'))
        reason = server_err.get('reason', '')
        e = ValueError(f"Failed to download the file. Status code: {response.status_code}, reason: {reason}")
        response.close()
        raise e


def upload_to_url(upload_url, file_path, headers, params, name):  # pragma: no cover
    try:
        from tqdm import tqdm
        from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
        has_progress_bar = True
    except ModuleNotFoundError:
        warnings.warn('You need to install tqdm and requests_toolbelt to see upload progress bar.')
        has_progress_bar = False

    if has_progress_bar:
        path = Path(file_path)
        total_size = path.stat().st_size
        fields = {}
        with tqdm(
                desc=f'Uploading {name}',
                total=total_size,
                unit="B",
                unit_scale=True,
                unit_divisor=1024,
        ) as bar:
            with open(file_path, "rb") as f:
                fields["file"] = ("dataset-sh-file", f)
                e = MultipartEncoder(fields=fields)
                m = MultipartEncoderMonitor(
                    e, lambda monitor: bar.update(monitor.bytes_read - bar.n)
                )
                headers["Content-Type"] = m.content_type
                response = requests.post(upload_url, data=m, headers=headers, params=params)


    else:
        with open(file_path, 'rb') as f:
            response = requests.post(
                upload_url,
                headers=headers,
                params=params,
                files={'file': ('data', f)}
            )
    response.raise_for_status()



import json

import click

from dataset_sh.clients.obj import LocalStorage
from dataset_sh.utils.misc import parse_tags


def _get_dataset_version(name, version, tag):
    dataset = LocalStorage().dataset(name)
    if version:
        dv = dataset.version(version)
    elif tag:
        dv = dataset.tag(tag)
    else:
        dv = dataset.latest()
    return dv


@click.group(name='local')
def local_cli():
    """to interact with dataset.sh local storage"""
    pass


@local_cli.command(name='import-as-latest')
@click.argument('name')
@click.argument('file')
@click.option('--tag', '-t', default='')
def import_(name, file, tag):
    """import dataset from file, and automatically tag it as latest"""
    click.echo(f'importing local file from {file}')
    LocalStorage().dataset(name).import_file(file_path=file, tags=parse_tags(tag))


@local_cli.command(name='import')
@click.argument('name')
@click.argument('file')
@click.option('--tag', '-t', default='latest')
def import_(name, file, tag):
    """import dataset from file and do not tag it as latest"""
    click.echo(f'importing local file from {file}')
    LocalStorage().dataset(name).import_file(file_path=file, tags=parse_tags(tag), as_latest=False)


@local_cli.command(name='meta')
@click.argument('name')
@click.option('--version', '-v', help='use this option to select version', default=None)
@click.option('--tag', '-t', help='use this option to select version with tag', default='latest')
def print_meta(name, version, tag):
    """print metadata of a dataset"""
    dv = _get_dataset_version(name, version, tag)
    if dv.exists():
        meta = dv.meta()
        if meta:
            click.echo(json.dumps(meta, indent=2))
    else:
        raise ValueError(f'{dv} do not exists.')


@local_cli.command(name='list-collections')
@click.argument('name')
@click.option('--version', '-v', help='use this option to select version', default=None)
@click.option('--tag', '-t', help='use this option to select version with tag', default='latest')
def list_collections(name, version, tag):
    """
    list collections of a dataset
    """
    dv = _get_dataset_version(name, version, tag)
    if dv.exists():
        meta = dv.meta_object()
        click.echo(f'Total collections: {len(meta.collections)}')
        for coll in meta.collections:
            click.echo(coll.name)
    else:
        raise ValueError(f'{dv} do not exists.')


@local_cli.command(name='print-code')
@click.argument('name')
@click.argument('collection_name')
@click.option('--version', '-v', help='use this option to select version', default=None)
@click.option('--tag', '-t', help='use this option to select version with tag', default='latest')
def print_code(name, collection_name, version, tag):
    dv = _get_dataset_version(name, version, tag)
    if dv.exists():
        code = dv.usage_code(collection_name)
        click.echo(code)
    else:
        raise ValueError(f'{dv} do not exists.')


@local_cli.command(name='print-sample')
@click.argument('name')
@click.argument('collection_name')
@click.option('--version', '-v', help='use this option to select version', default=None)
@click.option('--tag', '-t', help='use this option to select version with tag', default='latest')
def print_sample(name, collection_name, version, tag):
    dv = _get_dataset_version(name, version, tag)
    if dv.exists():
        samples = dv.sample(collection_name)
        click.echo(json.dumps(samples, indent=2))

    else:
        raise ValueError(f'{dv} do not exists.')


@local_cli.command(name='remove')
@click.argument('name')
@click.option('--force', '-f', default=False, help='Force remove dataset without confirmation.', is_flag=True)
def remove(name, force):
    """remove a managed dataset"""
    dataset = LocalStorage().dataset(name)
    do_remove = False
    if force:
        do_remove = True
    else:
        confirmation = click.prompt(f'Are you sure you want to remove all versions of dataset {name}? (y/N): ')
        if confirmation.lower() == 'y':
            do_remove = True

    if do_remove:
        dataset.delete()


@local_cli.command(name='remove-version')
@click.argument('name')
@click.option('--version', '-v', help='use this option to select version', default=None)
@click.option('--tag', '-t', help='use this option to select version with tag', default=None)
@click.option('--force', '-f', default=False, help='Force remove dataset without confirmation.', is_flag=True)
def remove_version(name, version, tag, force):
    """remove a managed dataset"""
    dv = _get_dataset_version(name, version, tag)
    do_remove = False
    if force:
        do_remove = True
    else:
        confirmation = click.prompt(f'Are you sure you want to remove all versions of dataset {name}? (y/N): ')
        if confirmation.lower() == 'y':
            do_remove = True

    if do_remove:
        dv.delete()


@local_cli.command(name='list')
@click.option('--namespace', '-n', help='select dataset store space to list.', default=None)
def list_datasets(namespace):
    """list datasets"""
    if namespace:
        items = LocalStorage().namespace(namespace).datasets()
    else:
        items = LocalStorage().datasets()

    click.echo(f'\nFound {len(items)} datasets:\n')
    items = sorted(items, key=lambda x: f'{x.namespace}/{x.dataset_name}')
    for item in items:
        click.echo(f'  {item.namespace}/{item.dataset_name}')
    click.echo('')


@local_cli.command(name='list-version')
@click.argument('name')
def list_dataset_versions(name):
    """list dataset versions"""
    dataset = LocalStorage().dataset(name)
    versions = dataset.versions()

    click.echo(f'\nFound {len(versions)} versions:\n')
    for item in versions:
        click.echo(f'  {item.version}')
    click.echo('')


@local_cli.command(name='tag')
@click.argument('name')
@click.argument('tag')
@click.argument('version')
def tag_dataset_version(name, tag, version):
    dataset = LocalStorage().dataset(name)
    dataset.set_tag(tag=tag, version=version)


@local_cli.command(name='untag')
@click.argument('name')
@click.argument('tag')
def untag_dataset_version(name, tag):
    dataset = LocalStorage().dataset(name)
    dataset.remove_tag(tag)


@local_cli.command(name='tag-info')
@click.argument('name')
@click.argument('tag')
def print_tag_info(name, tag):
    dataset = LocalStorage().dataset(name)
    tagged_version = dataset.resolve_tag(tag=tag)
    if tagged_version:
        click.echo(f"{tag} : {tagged_version}")
    else:
        click.echo(f"{tag} do not exists")



#!/usr/bin/env python3
import click
from dataset_sh.subcmds.remote import remote_cli
from dataset_sh.subcmds.local import local_cli
from dataset_sh.subcmds.project import project_cli
from dataset_sh.subcmds.profile import profile_cli
from dataset_sh.subcmds.file import file_cli


@click.group(name='dataset.sh')
def dsh():
    pass


cli = dsh
cli.add_command(remote_cli)
cli.add_command(local_cli)
cli.add_command(file_cli)
cli.add_command(profile_cli)
cli.add_command(project_cli)

if __name__ == '__main__':  # pragma: no cover
    cli()
